﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Hilbert
{
    class PLL
    {
        private int _points;
        private double[] _x;
        private double[] _y;
        private double _deltax;
        private double _ymin;
        private double _ymax;

        public PLL(PLL pll)
        {
            _points = pll.GetNumberOfPoints();
            _x = null;
            _y = null;
            _x = new double[_points];
            _y = new double[_points];
            _deltax = pll.GetDeltax();

            double[] x = pll.GetAllArgs();
            double[] y = pll.GetAllFunc();
            _ymin = y[0];
            _ymax = _ymin;
            for (int i = 0; i < x.Length; ++i)
            {
                _x[i] = x[i];
                _y[i] = y[i];
                if (_y[i] < _ymin) _ymin = _y[i];
                if (_y[i] > _ymax) _ymax = _y[i];
            }
        }

        public PLL(int points, double xmin, double xmax)
        {
            Initialize(points, xmin, xmax);
        }

        public double GetDerrivative(double x)
        {
            int low = (int)((x - _x[0]) / _deltax);
            if (low < 0) low = 0;
            if (low > _x.Length - 2) low = _x.Length - 2;
            return (_y[low + 1] - _y[low]) / _deltax;
        }

        public void Assign(double xmin, double xmax, double[] y)
        {
            if (y.Length < 3)
            {
                Console.WriteLine("Y array is too short");
                Environment.Exit(0);
            }
            if (xmin >= xmax)
            {
                Console.WriteLine("Assign: the limits are invalid");
                Environment.Exit(0);
            }
            _points = y.Length;
            _x = new double[_points];
            _y = new double[_points];
            _deltax = (xmax - xmin) / (_points - 1);
            _x[0] = xmin;
            _y[0] = y[0];
            _ymin = _y[0];
            _ymax = _ymin;
            for (int i = 1; i < _x.Length; ++i)
            {
                _x[i] = _x[i - 1] + _deltax;
                _y[i] = y[i];
                if (_y[i] < _ymin) _ymin = _y[i];
                if (_y[i] > _ymax) _ymax = _y[i];
            }
        }

        private void Initialize(int points, double xmin, double xmax)
        {
            if (points < 2)
            {
                Console.WriteLine("Number of blocks is too low");
                Environment.Exit(0);
            }
            if (xmin >= xmax)
            {
                Console.WriteLine("The limits are invalid");
                Environment.Exit(0);
            }
            _points = points;
            _x = new double[_points];
            _y = new double[_points];
            _deltax = (xmax - xmin) / (_points - 1);
            _x[0] = xmin;
            _y[0] = 0.0;
            for (int i = 1; i < _x.Length; ++i)
            {
                _x[i] = _x[i - 1] + _deltax;
                _y[i] = 0.0;
            }
            _ymin = _y[0];
            _ymax = _ymin;
        }

        private void Renormalize(double xmin, double xmax)
        {
            _deltax = (xmax - xmin) / (_points - 1);
            for (int i = 0; i < _x.Length; ++i)
            {
                _x[i] = xmin + i * _deltax;
                _y[i] = GetFunction(_x[i]);
            }
        }

        public void Update(double x, double delta, double mu)
        {
            if (x < _x[0])
            {
                Renormalize(x, _x[_x.Length - 1]);
            }

            if (x > _x[_x.Length - 1])
            {
                Renormalize(_x[0], x);
            }

            int left = (int)((x - _x[0]) / _deltax);
            if (left == _y.Length - 1)
            {
                _y[_y.Length - 1] += delta * mu;
                return;
            }

            double leftx = x - _x[left];
            double rightx = _x[left + 1] - x;
            _y[left + 1] += delta * leftx / _deltax * mu;
            _y[left] += delta * rightx / _deltax * mu;
            if (_y[left + 1] < _ymin) _ymin = _y[left + 1];
            if (_y[left + 1] > _ymax) _ymax = _y[left + 1];
            if (_y[left] < _ymin) _ymin = _y[left];
            if (_y[left] > _ymax) _ymax = _y[left];
        }

        public double GetFunction(double x)
        {
            if (x < _x[0])
            {
                double derrivative = (_y[1] - _y[0]) / _deltax;
                return _y[1] - derrivative * (_x[1] - x);
            }
            if (x > _x[_x.Length - 1])
            {
                double derrivative = (_y[_y.Length - 1] - _y[_y.Length - 2]) / _deltax;
                return _y[_y.Length - 2] + derrivative * (x - _x[_x.Length - 2]);
            }
            int left = (int)((x - _x[0]) / _deltax);
            if (left >= _x.Length - 1) return _y[left];
            double leftx = x - _x[left];
            return (_y[left + 1] - _y[left]) / _deltax * leftx + _y[left];
        }

        public void ShowData()
        {
            for (int i = 0; i < _x.Length; ++i)
            {
                Console.WriteLine("{0:0.0000} {1:0.0000}", _x[i], _y[i]);
            }
        }

        public int GetNumberOfPoints()
        {
            return _points;
        }

        public double GetY(int pos)
        {
            if (pos > _y.Length - 1) pos = _y.Length - 1;
            if (pos < 0) pos = 0;
            return _y[pos];
        }

        public double[] GetAllFunc()
        {
            return _y;
        }

        public double[] GetAllArgs()
        {
            return _x;
        }

        public double GetDeltax()
        {
            return _deltax;
        }

        public double GetYMin()
        {
            return _ymin;
        }

        public double GetYMax()
        {
            return _ymax;
        }

        public int GeNPoints()
        {
            return _points;
        }
    }
}
